@@ -4,7 +4,7 @@
  *              to build a moderately scalable and highly available server
  *              based on a cluster of servers, with more flexibility.
  *
- * Version:     $Id: tcp_vs.c,v 1.12 2003/02/22 04:51:57 wensong Exp $
+ * Version:     $Id: tcp_vs.c,v 1.15 2003/05/23 02:51:12 wensong Exp $
  *
  * Authors:     Wensong Zhang <wensong@linuxvirtualserver.org>
  *
@@ -15,7 +15,7 @@
  *
  */
 
-#define __KERNEL_SYSCALLS__             /*  for waitpid */
+#define __KERNEL_SYSCALLS__	/*  for waitpid */
 
 #include <linux/config.h>
 #include <linux/module.h>
@@ -51,7 +51,6 @@ EXPORT_SYMBOL(tcp_vs_xmit);
 EXPORT_SYMBOL(tcp_vs_recvbuffer);
 EXPORT_SYMBOL(tcp_vs_getword);
 EXPORT_SYMBOL(tcp_vs_getline);
-EXPORT_SYMBOL(tcp_vs_exp_exec);
 #ifdef CONFIG_TCP_VS_DEBUG
 EXPORT_SYMBOL(tcp_vs_get_debug_level);
 #endif
@@ -61,6 +60,9 @@ EXPORT_SYMBOL(tcp_vs_srvconn_get);
 EXPORT_SYMBOL(tcp_vs_srvconn_put);
 EXPORT_SYMBOL(tcp_vs_srvconn_new);
 EXPORT_SYMBOL(tcp_vs_srvconn_free);
+EXPORT_SYMBOL(tcp_vs_add_slowtimer);
+EXPORT_SYMBOL(tcp_vs_del_slowtimer);
+EXPORT_SYMBOL(tcp_vs_mod_slowtimer);
 
 
 struct tcp_vs_conn *
@@ -94,7 +96,8 @@ tcp_vs_conn_create(struct socket *sock, char *buffer, size_t buflen)
 }
 
 
-int tcp_vs_conn_release(struct tcp_vs_conn *conn)
+int
+tcp_vs_conn_release(struct tcp_vs_conn *conn)
 {
 	/* release the cloned socket */
 	sock_release(conn->csock);
@@ -129,7 +132,7 @@ skb_send_datagram_socket(const struct sk_buff *skb, struct socket *to)
 		return res;
 	written = res;
 
-	for (i=0; i<skb_shinfo(skb)->nr_frags; i++) {
+	for (i = 0; i < skb_shinfo(skb)->nr_frags; i++) {
 		char *vaddr;
 		skb_frag_t *frag = &skb_shinfo(skb)->frags[i];
 		struct page *page = frag->page;
@@ -143,13 +146,13 @@ skb_send_datagram_socket(const struct sk_buff *skb, struct socket *to)
 		kunmap(page);
 	}
 
-	for (list = skb_shinfo(skb)->frag_list; list; list=list->next) {
+	for (list = skb_shinfo(skb)->frag_list; list; list = list->next) {
 		res = skb_send_datagram_socket(list, to);
 		if (res < 0)
 			goto out;
 		written += res;
 	}
-  out:
+      out:
 	return written;
 }
 
@@ -177,7 +180,8 @@ skb_entail(struct sock *sk, struct tcp_opt *tp, struct sk_buff *skb)
 		tp->send_head = skb;
 }
 
-static int tcp_vs_relay_socket(struct socket *from, struct socket *to)
+static int
+tcp_vs_relay_socket(struct socket *from, struct socket *to)
 {
 	struct sk_buff *skb;
 	int len;
@@ -211,9 +215,9 @@ static int tcp_vs_relay_socket(struct socket *from, struct socket *to)
 				kfree_skb(skb);
 				return written;
 			}
-		  sendagain:
-			res = tcp_vs_sendbuffer(to, skb->data+written,
-						skb->len-written, 0);
+		      sendagain:
+			res = tcp_vs_sendbuffer(to, skb->data + written,
+						skb->len - written, 0);
 			if (res < 0) {
 				kfree_skb(skb);
 				return written;
@@ -230,7 +234,7 @@ static int tcp_vs_relay_socket(struct socket *from, struct socket *to)
 	/* we cannot release the skb here, but we do need to call
 	   its destructor so that the sock_rfree can update the
 	   source sk->rmem_alloc correctly. */
-	if(skb->destructor) {
+	if (skb->destructor) {
 		skb->destructor(skb);
 	}
 
@@ -239,7 +243,7 @@ static int tcp_vs_relay_socket(struct socket *from, struct socket *to)
 	skb->dst = NULL;
 #ifdef CONFIG_NETFILTER_DEBUG
 	skb->nf_debug = 0;
-#endif /* CONFIG_NETFILTER_DEBUG */
+#endif				/* CONFIG_NETFILTER_DEBUG */
 
 	sk = to->sk;
 	tp = &(sk->tp_pinfo.af_tcp);
@@ -266,7 +270,8 @@ static int tcp_vs_relay_socket(struct socket *from, struct socket *to)
  *   between the tcpvs and the selected server. Terminate until that
  *   the two connections are done.
  */
-int tcp_vs_conn_handle(struct tcp_vs_conn *conn, struct tcp_vs_service *svc)
+int
+tcp_vs_conn_handle(struct tcp_vs_conn *conn, struct tcp_vs_service *svc)
 {
 	struct socket *csock, *dsock;
 	DECLARE_WAITQUEUE(wait1, current);
@@ -285,21 +290,21 @@ int tcp_vs_conn_handle(struct tcp_vs_conn *conn, struct tcp_vs_service *svc)
 
 	/*
 	   call its scheduler to process the connect request, the scheduler may:
-	     1, Select a destination server and return 0, or;
-	     2, Deal with the requestion alone and return 1, or;
-	     3, Return -1  when could not find a right server
-	     4, return -2 when other errors such as socket broken occur.
-	*/
+	   1, Select a destination server and return 0, or;
+	   2, Deal with the requestion alone and return 1, or;
+	   3, Return -1  when could not find a right server
+	   4, return -2 when other errors such as socket broken occur.
+	 */
 	switch (svc->scheduler->schedule(conn, svc)) {
-	case 1:         /* scheduler has done all the work */
+	case 1:		/* scheduler has done all the work */
 		return 0;
 
-	case 0:         /* further process needed */
+	case 0:		/* further process needed */
 		break;
-	case -1:	/* try to redirect the connection to other sockets */
+	case -1:		/* try to redirect the connection to other sockets */
 		if (svc->conf.redirect_port) {
 			redirect_to_local(conn, svc->conf.redirect_addr,
-			svc->conf.redirect_port);
+					  svc->conf.redirect_port);
 			return 0;
 		}
 		TCP_VS_ERR("no destination available\n");
@@ -333,13 +338,13 @@ int tcp_vs_conn_handle(struct tcp_vs_conn *conn, struct tcp_vs_service *svc)
 			break;
 
 		/* Do we have data from server? */
-		if(!skb_queue_empty(&(dsock->sk->receive_queue))) {
+		if (!skb_queue_empty(&(dsock->sk->receive_queue))) {
 			if (tcp_vs_relay_socket(dsock, csock) == 0)
 				break;
 		}
 
 		/* Do we have data from client? */
-		if(!skb_queue_empty(&(csock->sk->receive_queue))) {
+		if (!skb_queue_empty(&(csock->sk->receive_queue))) {
 			if (tcp_vs_relay_socket(csock, dsock) == 0)
 				break;
 		}
@@ -388,9 +393,9 @@ enum {
 
 struct tcp_vs_child_table {
 	struct tcp_vs_child children[KTCPVS_CHILD_HARD_LIMIT];
-	int		max_daemons_limit;
-	int		idle_spawn_rate;
-	unsigned long	last_modified;	/* last time of add/killing child */
+	int max_daemons_limit;
+	int idle_spawn_rate;
+	unsigned long last_modified;	/* last time of add/killing child */
 };
 
 static int tcp_vs_child(void *__child);
@@ -415,7 +420,8 @@ kill_child(struct tcp_vs_child_table *tbl, int slot)
 	tbl->last_modified = jiffies;
 }
 
-static inline void update_child_status(struct tcp_vs_child *chd, __u16 status)
+static inline void
+update_child_status(struct tcp_vs_child *chd, __u16 status)
 {
 	chd->status = status;
 }
@@ -435,8 +441,7 @@ child_pool_maintenance(struct tcp_vs_child_table *tbl,
 		int status;
 
 		if (i >= tbl->max_daemons_limit
-		    && free_length == tbl->idle_spawn_rate)
-			break;
+		    && free_length == tbl->idle_spawn_rate) break;
 		status = tbl->children[i].status;
 		switch (status) {
 		case SERVER_DEAD:
@@ -468,10 +473,11 @@ child_pool_maintenance(struct tcp_vs_child_table *tbl,
 	} else if (idle_count < svc->conf.minSpareServers) {
 		if (free_length) {
 			if (tbl->idle_spawn_rate > 8 && net_ratelimit())
-				TCP_VS_INFO("Server %s seems busy, you may "
-					    "need to increase StartServers, "
-					    "or Min/MaxSpareServers\n",
-					    svc->ident.name);
+				TCP_VS_INFO
+				    ("Server %s seems busy, you may "
+				     "need to increase StartServers, "
+				     "or Min/MaxSpareServers\n",
+				     svc->ident.name);
 			/* spawn a batch of children */
 			for (i = 0; i < free_length; i++)
 				make_child(tbl, free_slots[i], svc);
@@ -479,30 +485,32 @@ child_pool_maintenance(struct tcp_vs_child_table *tbl,
 			if (tbl->idle_spawn_rate < MAX_SPAWN_RATE)
 				tbl->idle_spawn_rate *= 2;
 		} else if (net_ratelimit())
-			TCP_VS_INFO("Server %s reached MaxClients setting, "
-				    "consider raising the MaxClients "
-				    "setting\n", svc->ident.name);
+			TCP_VS_INFO
+			    ("Server %s reached MaxClients setting, "
+			     "consider raising the MaxClients "
+			     "setting\n", svc->ident.name);
 	} else {
 		/* if the number of spare servers remains in the interval
 		   (minSpareServers, maxSpareServers] and the time of
 		   last modified is larger than ten minutes, we try to
 		   kill one spare child in order to release some resource. */
 		if (idle_count > svc->conf.minSpareServers
-		    && jiffies - tbl->last_modified > 600*HZ)
+		    && jiffies - tbl->last_modified > 600 * HZ)
 			kill_child(tbl, to_kill);
 		tbl->idle_spawn_rate = 1;
 	}
 }
 
 
-static int tcp_vs_child(void *__child)
+static int
+tcp_vs_child(void *__child)
 {
 	struct tcp_vs_conn *conn;
 	struct socket *sock;
 	int ret = 0;
 	char *Buffer;
 	size_t BufLen;
-	struct tcp_vs_child *chd = (struct tcp_vs_child *)__child;
+	struct tcp_vs_child *chd = (struct tcp_vs_child *) __child;
 	struct tcp_vs_service *svc = chd->svc;
 
 	/* DECLARE_WAIT_QUEUE_HEAD(queue); */
@@ -521,7 +529,8 @@ static int tcp_vs_child(void *__child)
 
 	/* Block all signals except SIGKILL and SIGSTOP */
 	spin_lock_irq(&current->sigmask_lock);
-	siginitsetinv(&current->blocked, sigmask(SIGKILL) | sigmask(SIGSTOP));
+	siginitsetinv(&current->blocked,
+		      sigmask(SIGKILL) | sigmask(SIGSTOP));
 	recalc_sigpending(current);
 	spin_unlock_irq(&current->sigmask_lock);
 
@@ -532,7 +541,7 @@ static int tcp_vs_child(void *__child)
 		goto out;
 	}
 
-	Buffer = (char*) get_free_page(GFP_KERNEL);
+	Buffer = (char *) get_free_page(GFP_KERNEL);
 	if (Buffer == NULL) {
 		ret = -ENOMEM;
 		goto out;
@@ -586,8 +595,8 @@ static int tcp_vs_child(void *__child)
 		atomic_dec(&svc->conns);
 	}
 
-	free_page((unsigned long)Buffer);
-  out:
+	free_page((unsigned long) Buffer);
+      out:
 	update_child_status(chd, SERVER_DEAD);
 	atomic_dec(&svc->childcount);
 	LeaveFunction(3);
@@ -595,11 +604,12 @@ static int tcp_vs_child(void *__child)
 }
 
 
-static int tcp_vs_daemon(void *__svc)
+static int
+tcp_vs_daemon(void *__svc)
 {
 	int waitpid_result;
 	int i;
-	struct tcp_vs_service *svc = (struct tcp_vs_service *)__svc;
+	struct tcp_vs_service *svc = (struct tcp_vs_service *) __svc;
 	struct tcp_vs_child_table *child_table = NULL;
 
 	DECLARE_WAIT_QUEUE_HEAD(WQ);
@@ -613,12 +623,14 @@ static int tcp_vs_daemon(void *__svc)
 
 	/* Block all signals except SIGKILL and SIGSTOP */
 	spin_lock_irq(&current->sigmask_lock);
-	siginitsetinv(&current->blocked, sigmask(SIGKILL) | sigmask(SIGSTOP) );
+	siginitsetinv(&current->blocked,
+		      sigmask(SIGKILL) | sigmask(SIGSTOP));
 	recalc_sigpending(current);
 	spin_unlock_irq(&current->sigmask_lock);
 
 	if (!svc->scheduler) {
-		TCP_VS_ERR("%s's scheduler is not bound\n", svc->ident.name);
+		TCP_VS_ERR("%s's scheduler is not bound\n",
+			   svc->ident.name);
 		goto out;
 	}
 
@@ -637,7 +649,7 @@ static int tcp_vs_daemon(void *__svc)
 	memset(child_table, 0, sizeof(*child_table));
 	child_table->idle_spawn_rate = 1;
 
-	for (i=0; i<svc->conf.startservers; i++)
+	for (i = 0; i < svc->conf.startservers; i++)
 		make_child(child_table, i, svc);
 
 	/* Then wait for deactivation */
@@ -649,7 +661,7 @@ static int tcp_vs_daemon(void *__svc)
 		child_pool_maintenance(child_table, svc);
 
 		/* reap the zombie daemons */
-		waitpid_result = waitpid(-1, NULL, __WCLONE|WNOHANG);
+		waitpid_result = waitpid(-1, NULL, __WCLONE | WNOHANG);
 	}
 
 	/* Wait for tcp_vs_child to stop, one second per iteration */
@@ -659,12 +671,12 @@ static int tcp_vs_daemon(void *__svc)
 	/* reap the zombie daemons */
 	waitpid_result = 1;
 	while (waitpid_result > 0)
-		waitpid_result = waitpid(-1, NULL, __WCLONE|WNOHANG);
+		waitpid_result = waitpid(-1, NULL, __WCLONE | WNOHANG);
 
 	/* stop listening */
 	StopListening(svc);
 
-  out:
+      out:
 	if (child_table)
 		vfree(child_table);
 	svc->start = 0;
@@ -675,7 +687,8 @@ static int tcp_vs_daemon(void *__svc)
 }
 
 
-static int master_daemon(void *unused)
+static int
+master_daemon(void *unused)
 {
 	int waitpid_result;
 	struct list_head *l;
@@ -686,20 +699,21 @@ static int master_daemon(void *unused)
 
 	MOD_INC_USE_COUNT;
 
-	sprintf(current->comm,"ktcpvs master");
+	sprintf(current->comm, "ktcpvs master");
 	lock_kernel();
 	daemonize();
 
 	/* Block all signals except SIGKILL and SIGSTOP */
 	spin_lock_irq(&current->sigmask_lock);
-	siginitsetinv(&current->blocked, sigmask(SIGKILL) | sigmask(SIGSTOP) );
+	siginitsetinv(&current->blocked,
+		      sigmask(SIGKILL) | sigmask(SIGSTOP));
 	recalc_sigpending(current);
 	spin_unlock_irq(&current->sigmask_lock);
 
 	/* main loop */
 	while (sysctl_ktcpvs_unload == 0) {
 		read_lock(&__tcp_vs_svc_lock);
-		list_for_each (l, &tcp_vs_svc_list) {
+		list_for_each(l, &tcp_vs_svc_list) {
 			svc = list_entry(l, struct tcp_vs_service, list);
 			if (!atomic_read(&svc->running) && svc->start)
 				kernel_thread(tcp_vs_daemon, svc, 0);
@@ -716,7 +730,7 @@ static int master_daemon(void *unused)
 		interruptible_sleep_on_timeout(&WQ, HZ);
 
 		/* reap the daemons */
-		waitpid_result = waitpid(-1, NULL, __WCLONE|WNOHANG);
+		waitpid_result = waitpid(-1, NULL, __WCLONE | WNOHANG);
 	}
 
 	/* Wait for tcp_vs daemons to stop, one second per iteration */
@@ -726,7 +740,7 @@ static int master_daemon(void *unused)
 	/* reap the zombie daemons */
 	waitpid_result = 1;
 	while (waitpid_result > 0)
-		waitpid_result = waitpid(-1, NULL, __WCLONE|WNOHANG);
+		waitpid_result = waitpid(-1, NULL, __WCLONE | WNOHANG);
 
 	/* flush all the virtual servers */
 	tcp_vs_flush();
@@ -740,7 +754,8 @@ static int master_daemon(void *unused)
 }
 
 
-static int __init ktcpvs_init(void)
+static int __init
+ktcpvs_init(void)
 {
 	tcp_vs_control_start();
 
@@ -748,7 +763,7 @@ static int __init ktcpvs_init(void)
 
 	tcp_vs_srvconn_init();
 
-	(void)kernel_thread(master_daemon, NULL, 0);
+	(void) kernel_thread(master_daemon, NULL, 0);
 
 	TCP_VS_INFO("ktcpvs loaded.\n");
 
@@ -756,7 +771,8 @@ static int __init ktcpvs_init(void)
 }
 
 
-static void __exit ktcpvs_cleanup(void)
+static void __exit
+ktcpvs_cleanup(void)
 {
 	tcp_vs_srvconn_cleanup();
 
