@@ -11,7 +11,7 @@
 
 #include <asm/types.h>		/* For __uXX types */
 
-#define TCP_VS_VERSION_CODE		0x00000D
+#define TCP_VS_VERSION_CODE		0x00000E
 #define NVERSION(version)                       \
 	(version >> 16) & 0xFF,                 \
 	(version >> 8) & 0xFF,                  \
@@ -400,54 +400,50 @@ struct tcp_vs_child {
 	volatile __u16 status;	/* child status */
 };
 
-/*
- *	HTTP Method
+/**
+ * @defgroup Methods List of Methods recognized by the server
+ * @{
  */
-typedef enum http_method_s {
-	HTTP_M_UNKNOWN,
-	HTTP_M_OPTIONS,
-	HTTP_M_GET,
-	HTTP_M_HEAD,
-	HTTP_M_POST,
-	HTTP_M_PUT,
-	HTTP_M_DELETE,
-	HTTP_M_TRACE,
-	HTTP_M_MAX
-} http_method;
+#define HTTP_M_UNKNOWN		     -1
+#define HTTP_M_GET                   0       /* RFC 2616: HTTP */
+#define HTTP_M_HEAD		     1
+#define HTTP_M_PUT                   2       /*  :             */
+#define HTTP_M_POST                  3
+#define HTTP_M_DELETE                4
+#define HTTP_M_CONNECT               5
+#define HTTP_M_OPTIONS               6
+#define HTTP_M_TRACE                 7       /* RFC 2616: HTTP */
+#define HTTP_M_PATCH                 8       /* no rfc(!)  ### remove this one? */
+#define HTTP_M_PROPFIND              9       /* RFC 2518: WebDAV */
+#define HTTP_M_PROPPATCH             10       /*  :               */
+#define HTTP_M_MKCOL                 11
+#define HTTP_M_COPY                  12
+#define HTTP_M_MOVE                  13
+#define HTTP_M_LOCK                  14
+#define HTTP_M_UNLOCK                15      /* RFC 2518: WebDAV */
+#define HTTP_M_VERSION_CONTROL       16      /* RFC 3253: WebDAV Versioning */
+#define HTTP_M_CHECKOUT              17      /*  :                          */
+#define HTTP_M_UNCHECKOUT            18
+#define HTTP_M_CHECKIN               19
+#define HTTP_M_UPDATE                20
+#define HTTP_M_LABEL                 21
+#define HTTP_M_REPORT                22
+#define HTTP_M_MKWORKSPACE           23
+#define HTTP_M_MKACTIVITY            24
+#define HTTP_M_BASELINE_CONTROL      25
+#define HTTP_M_MERGE                 26
+#define HTTP_M_INVALID               27      /* RFC 3253: WebDAV Versioning */
 
-/*
- *	HTTP Version
- */
-typedef enum http_version_s {
-	HTTP_V_UNKNOWN,
-	HTTP_V_0_9,
-	HTTP_V_1_0,
-	HTTP_V_1_1
-} http_version;
-
-/*
- *	HTTP MIME header to be parsed
- */
-enum {
-	TRANSFER_ENCODING = 0,
-	CONTENT_LENGTH,
-	CONNECTION,
-	CONTENT_TYPE,
-	SET_COOKIE,
-	SET_COOKIE2,
-	COOKIE,
-	MAX_PARSER_ID
-};
+#define HTTP_VERSION(major,minor) (1000*(major)+(minor))
 
 /* cookie components that act as key */
 typedef struct cookie_key_s {
 	char *name;
-	char *value;
 	ulong sid;
 } cookie_key_t;
 
 /* HTTP cookie */
-typedef struct http_cookie {
+typedef struct http_cookie_s {
 	struct list_head c_list;
 	int discard;
 	unsigned int max_age;
@@ -472,7 +468,7 @@ typedef struct http_request_s {
 	unsigned int parsed_len;
 
 	/* request method */
-	http_method method;
+	int method;
 	const char *method_str;
 
 	/* request URI */
@@ -481,7 +477,7 @@ typedef struct http_request_s {
 	unsigned int method_len;
 
 	/* http version */
-	http_version version;
+	int version;
 	const char *version_str;
 	unsigned int version_len;
 
@@ -500,16 +496,24 @@ typedef struct http_response_s {
 	http_mime_header_t mime;
 } http_response_t;
 
+/* buffer to store http lines */
+typedef struct http_buf_s {
+	struct list_head	b_list;
+	char			*buf;
+	int			data_len;
+} http_buf_t;
+
 /*
  *  control block to read data from socket
  */
 typedef struct http_read_ctl_block_s {
 	struct socket *sock;	/* socket that message read from */
-	char *buffer;		/* buffer to store read message */
-	int len;		/* size of buffer */
+	struct list_head	buf_entry_list; /* buffer list */
+	http_buf_t		*cur_buf;	/* current buffer */
 	char *info;		/* point to the current information */
 	int offset;		/* offset of remaining bytes */
 	int remaining;		/* remaining bytes not return */
+	int			buf_size;	/* buffer size */
 	int flag;		/* read flag */
 } http_read_ctl_block_t;
 
@@ -578,7 +582,7 @@ extern void tcp_vs_srvconn_free(server_conn_t * sc);
 extern int tcp_vs_srvconn_init(void);
 extern void tcp_vs_srvconn_cleanup(void);
 
-/* from ip_vs_timer.c */
+/* from tcp_vs_timer.c */
 void assert_slowtimer(int pos);
 extern void tcp_vs_add_slowtimer(slowtimer_t * timer);
 extern int tcp_vs_del_slowtimer(slowtimer_t * timer);
@@ -608,18 +612,25 @@ extern int parse_http_status_line(char *buffer, size_t len,
 
 extern void http_mime_parser_init(void);
 
-extern void http_mime_parse_enable(int parser_id, int enable);
-
 extern int http_mime_parse(char *buffer, int len,
 			   http_mime_header_t * mime);
 
+extern char* search_sep(const char *s, int len, const char *sep);
+
+extern long get_chunk_size(char *b);
+
+/* from tcp_vs_http_trans.c */
 extern int relay_http_message_body(struct socket *dsock,
 				   http_read_ctl_block_t * ctl_blk,
 				   http_mime_header_t * mime);
 
+extern int http_read_init(http_read_ctl_block_t *ctl_blk, struct socket *sock);
+
+extern void http_read_free(http_read_ctl_block_t *read_ctl);
+
 extern int data_available(http_read_ctl_block_t * ctl_blk);
 
-extern int http_read_line(http_read_ctl_block_t * ctl_blk);
+extern int http_read_line(http_read_ctl_block_t * ctl_blk, int grow);
 
 #endif				/* __KERNEL__ */
 
