@@ -7,7 +7,7 @@
  * tcp_vs_chttp.c: KTCPVS content-based scheduling module for HTTP service
  *		   with cookie support.
  *
- * Version:	$Id: tcp_vs_chttp.c,v 1.3 2003/05/28 10:26:15 wensong Exp $
+ * Version:	$Id: tcp_vs_chttp.c,v 1.6 2003/06/14 09:04:11 wensong Exp $
  *
  * Authors:     Wensong Zhang, <wensong@linuxvirtualserver.org>
  *              Hai Long, <david_lung@yahoo.com>
@@ -47,20 +47,6 @@ typedef struct session_entry_s {
 	unsigned int ref_cnt;
 } session_entry_t;
 
-/* buffer to store http message header */
-typedef struct http_message_header_buf_s {
-	struct list_head b_list;
-	char *buf;
-	int buf_size;
-} http_message_header_buf_t;
-
-/* buffer controll structure for http message headers */
-typedef struct http_header_buf_ctl_s {
-	struct list_head buf_entry_list;
-	http_message_header_buf_t *cur_buf;	/* current buffer */
-	int offset;		/* offset in current buffer */
-} http_header_buf_ctl_t;
-
 /* lock for session id */
 static spinlock_t session_id_lock = SPIN_LOCK_UNLOCKED;
 
@@ -168,9 +154,9 @@ inject_session_id_cookie(struct socket *sock, int set_cookie2)
 	char buf[80];		/* avoid kmalloc */
 	ulong id;
 
-	spin_lock(session_id_lock);
+	spin_lock(&session_id_lock);
 	id = ktcpvs_session_id++;
-	spin_unlock(session_id_lock);
+	spin_unlock(&session_id_lock);
 
 	if (set_cookie2) {
 		sprintf(buf,
@@ -248,7 +234,6 @@ free_cookie_list(struct list_head *cookie_list)
 		/* free this cookie */
 		if (cookie->key != NULL) {
 			kfree(cookie->key->name);
-			kfree(cookie->key->value);
 			kfree(cookie->key);
 		}
 		kfree(cookie);
@@ -270,9 +255,9 @@ find_server_by_session_id(ulong sid)
 	EnterFunction(6);
 
 	se.sid = sid;
-	spin_lock(avl_tbl_lock);
+	spin_lock(&avl_tbl_lock);
 	session_entry = avl_find(session_tbl, &se);
-	spin_unlock(avl_tbl_lock);
+	spin_unlock(&avl_tbl_lock);
 
 	if (session_entry != NULL) {
 		dest = session_entry->dest;
@@ -339,7 +324,7 @@ cookie_expire(unsigned long data)
 	se.sid = sid;
 
 	/* reduce the reference count in session table */
-	spin_lock(avl_tbl_lock);
+	spin_lock(&avl_tbl_lock);
 	session_entry = avl_find(session_tbl, &se);
 	if (session_entry != NULL) {
 		session_entry->ref_cnt--;
@@ -357,7 +342,6 @@ cookie_expire(unsigned long data)
 	spin_unlock(&avl_tbl_lock);
 
 	kfree(cookie_entry->key->name);
-	kfree(cookie_entry->key->value);
 	kfree(cookie_entry->key);
 	kfree(cookie_entry);
 
@@ -372,30 +356,21 @@ cookie_expire(unsigned long data)
 static cookie_entry_t *
 new_cookie_entry(http_cookie_t * cookie)
 {
-	cookie_entry_t *cookie_entry;
+	cookie_entry_t *ce;
 
 	EnterFunction(6);
 
-	assert (cookie != NULL);	
-	assert (cookie->max_age != 0);
-	cookie_entry =
-	    (cookie_entry_t *) kmalloc(sizeof(cookie_entry_t), GFP_KERNEL);
-	if (cookie_entry == NULL) {
-		goto out;
+	assert(cookie != NULL);
+	assert(cookie->max_age != 0);
+
+	ce = (cookie_entry_t *) kmalloc(sizeof(cookie_entry_t), GFP_ATOMIC);
+	if (ce != NULL) {
+		ce->key = cookie->key;
+		cookie->key = NULL;	/* avoid to be freed by free_cookie_list */
 	}
 
-	memset(cookie_entry, 0, sizeof(cookie_entry_t));
-
-	cookie_entry->key = cookie->key;
-	cookie->key = NULL;	/* avoid to be freed by free_cookie_list */
-
-	init_slowtimer(&cookie_entry->cookie_expire_timer);
-	cookie_entry->cookie_expire_timer.data		= (unsigned long) cookie_entry;
-	cookie_entry->cookie_expire_timer.function = cookie_expire;
-	cookie_entry->cookie_expire_timer.expires	= jiffies + cookie->max_age * HZ;
-      out:
 	LeaveFunction(6);
-	return cookie_entry;
+	return ce;
 }
 
 
@@ -406,23 +381,17 @@ new_cookie_entry(http_cookie_t * cookie)
 static void
 update_cookie_entry(cookie_entry_t * cookie_entry, http_cookie_t * cookie)
 {
-	char *v;
 
 	EnterFunction(6);
 
-	/* replace the cookie value with new value,
-	   the old value string will be freed later */
-	v = cookie_entry->key->value;
-	cookie_entry->key->value = cookie->key->value;
-	cookie->key->value = v;
-
 	if (cookie->discard == 1) {
 		cookie->max_age =
 		    MIN(COOKIE_DISCARD_TIME, cookie->max_age);
 	}
 
 	tcp_vs_del_slowtimer(&cookie_entry->cookie_expire_timer);
-	cookie_entry->cookie_expire_timer.expires = jiffies + cookie->max_age * HZ;
+	cookie_entry->cookie_expire_timer.expires =
+	    jiffies + cookie->max_age * HZ;
 	tcp_vs_add_slowtimer(&cookie_entry->cookie_expire_timer);
 
 	LeaveFunction(6);
@@ -440,7 +409,6 @@ free_cookie_entry(void *avl_item, void *avl_param)
 
 	tcp_vs_del_slowtimer(&cookie_entry->cookie_expire_timer);
 	kfree(cookie_entry->key->name);
-	kfree(cookie_entry->key->value);
 	kfree(cookie_entry->key);
 }
 
@@ -481,7 +449,7 @@ http_set_cookie_handler(struct list_head *cookie_list,
 			struct tcp_vs_dest *dest, ulong sid)
 {
 	struct list_head *l;
-	cookie_entry_t *cookie_entry, *ce;
+	cookie_entry_t *dup, *ce;
 	session_entry_t *session_entry, *se;
 	http_cookie_t *cookie = NULL;
 	int ret = -1;
@@ -502,7 +470,7 @@ http_set_cookie_handler(struct list_head *cookie_list,
 	session_entry->dest = dest;
 	session_entry->ref_cnt = 0;
 
-	spin_lock(avl_tbl_lock);
+	spin_lock(&avl_tbl_lock);
 
 	se = avl_insert(session_tbl, session_entry);
 	if (se != NULL) {	/* duplicate session entry */
@@ -514,29 +482,34 @@ http_set_cookie_handler(struct list_head *cookie_list,
 	list_for_each(l, cookie_list) {
 		cookie = list_entry(l, http_cookie_t, c_list);
 		cookie->key->sid = sid;
-		assert (cookie->key->name != NULL);
+		assert(cookie->key->name != NULL);
 
-		cookie_entry = new_cookie_entry(cookie);
-		if (cookie_entry == NULL) {
+		ce = new_cookie_entry(cookie);
+		if (ce == NULL) {
 			TCP_VS_ERR("Out of memory!\n");
 			goto out;
 		}
 
-		ce = avl_insert(cookie_tbl, cookie_entry);
-		if (ce != NULL) {
+		dup = avl_insert(cookie_tbl, ce);
+		if (dup != NULL) {
 			/* update cookie entry */
-			update_cookie_entry(ce, cookie);
-			kfree(cookie_entry);
+			update_cookie_entry(dup, cookie);
+			kfree(ce);
 		} else {	/* successfully inserted */
 			session_entry->ref_cnt++;
-			tcp_vs_add_slowtimer(&cookie_entry->cookie_expire_timer);
+			init_slowtimer(&ce->cookie_expire_timer);
+			ce->cookie_expire_timer.data = (unsigned long) ce;
+			ce->cookie_expire_timer.function = cookie_expire;
+			ce->cookie_expire_timer.expires =
+			    jiffies + cookie->max_age * HZ;
+			tcp_vs_add_slowtimer(&ce->cookie_expire_timer);
 		}
 	}			/* end of list_for_each */
 
 	ret = 0;
       out:
 	free_cookie_list(cookie_list);
-	spin_unlock(avl_tbl_lock);
+	spin_unlock(&avl_tbl_lock);
 	LeaveFunction(6);
 	return ret;
 }
@@ -551,6 +524,7 @@ chttp_get_response(struct socket *csock, server_conn_t * sc,
 		   int buflen, int *close)
 {
 	http_read_ctl_block_t read_ctl_blk;
+	http_buf_t	buff;	
 	http_response_t resp;
 	int len, ret = -1;
 	struct socket *dsock = sc->sock;
@@ -559,10 +533,16 @@ chttp_get_response(struct socket *csock, server_conn_t * sc,
 
 	EnterFunction(5);
 
+	INIT_LIST_HEAD(&buff.b_list);
+	buff.buf = buffer;
+	buff.data_len = 0;
+
 	memset(&read_ctl_blk, 0, sizeof(read_ctl_blk));
-	read_ctl_blk.buffer = buffer;
-	read_ctl_blk.len = buflen;
+	INIT_LIST_HEAD(&read_ctl_blk.buf_entry_list);
+	read_ctl_blk.cur_buf  = &buff;
+	read_ctl_blk.buf_size = buflen;
 	read_ctl_blk.sock = dsock;
+	list_add (&buff.b_list, &read_ctl_blk.buf_entry_list);
 
 	*close = 0;
 
@@ -572,7 +552,7 @@ chttp_get_response(struct socket *csock, server_conn_t * sc,
 	}
 
 	/* read status line from server */
-	len = http_read_line(&read_ctl_blk);
+	len = http_read_line(&read_ctl_blk, 0);
 	if (len < 0) {
 		TCP_VS_ERR("Error in reading status line from server\n");
 		goto exit;
@@ -591,11 +571,9 @@ chttp_get_response(struct socket *csock, server_conn_t * sc,
 		goto exit;
 	}
 
-	http_mime_parse_enable(CONTENT_TYPE, resp.status_code == 206);
-
 	/* parse MIME header */
 	do {
-		if ((len = http_read_line(&read_ctl_blk)) < 0) {
+		if ((len = http_read_line(&read_ctl_blk, 0)) < 0) {
 			goto exit;
 		}
 
@@ -603,11 +581,10 @@ chttp_get_response(struct socket *csock, server_conn_t * sc,
 		/*inject a cookie with session id at the end of the http header */
 		if ((len == 0) && resp.mime.cookie) {
 			sid = req->mime.session_id;
-			if (sid == 0) {
-				sid =
-				    inject_session_id_cookie(csock,
-							     resp.
-							     mime.set_cookie2);
+			if ((sid == 0) || (sid > ktcpvs_session_id)) {
+				sid = inject_session_id_cookie(csock,
+							       resp.mime.
+							       set_cookie2);
 			}
 		}
 
@@ -662,150 +639,57 @@ chttp_get_response(struct socket *csock, server_conn_t * sc,
 }
 
 
-/****************************************************************************
-*	Init http header buffer control structure
-*/
-static int
-http_header_buffer_init(http_header_buf_ctl_t * buf_ctl)
-{
-	int ret = -1;
-	http_message_header_buf_t *buf_entry;
-
-	INIT_LIST_HEAD(&buf_ctl->buf_entry_list);
-	buf_ctl->offset = 0;
-	buf_entry = (http_message_header_buf_t *)
-	    kmalloc(sizeof(http_message_header_buf_t), GFP_KERNEL);
-	if (buf_entry != NULL) {
-		buf_entry->buf = (char *) get_free_page(GFP_KERNEL);
-		if (buf_entry->buf != NULL) {
-			buf_entry->buf_size = PAGE_SIZE;
-			list_add_tail(&buf_entry->b_list,
-				      &buf_ctl->buf_entry_list);
-			buf_ctl->cur_buf = buf_entry;
-			ret = 0;
-		}
-	}
-
-	return ret;
-}
-
-
-/****************************************************************************
-*	Free http header buffer
-*/
-static void
-http_header_buffer_free(http_header_buf_ctl_t * buf_ctl)
-{
-	struct list_head *l, *temp;
-	http_message_header_buf_t *buf_entry;
-
-	list_for_each_safe(l, temp, &buf_ctl->buf_entry_list) {
-		list_del(l);
-		buf_entry =
-		    list_entry(l, http_message_header_buf_t, b_list);
-		free_page((unsigned long) buf_entry->buf);
-		kfree(buf_entry);
-	}
-
-	return;
-}
-
-
-/****************************************************************************
-*	copy data to http header buffer
-*/
-static int
-write_buffer(http_header_buf_ctl_t * buf_ctl, char *data, int len)
-{
-
-	http_message_header_buf_t *buf_entry;
-	int count, remaining = len, ret = -1;
-	int spare_space = buf_ctl->cur_buf->buf_size - buf_ctl->offset;
-	char *pos = data;
-
-	EnterFunction(6);
-
-      write_again:
-	count = MIN(remaining, spare_space);
-	memcpy(buf_ctl->cur_buf->buf + buf_ctl->offset, pos, count);
-	remaining -= count;
-	if (remaining != 0) {
-		buf_entry = (http_message_header_buf_t *)
-		    kmalloc(sizeof(http_message_header_buf_t), GFP_KERNEL);
-		if (buf_entry != NULL) {
-			buf_entry->buf =
-			    (char *) get_free_page(GFP_KERNEL);
-			if (buf_entry->buf != NULL) {
-				buf_entry->buf_size = PAGE_SIZE;
-				buf_ctl->offset = 0;
-				list_add_tail(&buf_entry->b_list,
-					      &buf_ctl->buf_entry_list);
-				buf_ctl->cur_buf = buf_entry;
-				spare_space = PAGE_SIZE;
-				pos += count;
-				goto write_again;
-			}
-		}
-	} else {
-		buf_ctl->offset += count;
-		ret = 0;
-	}
-
-	LeaveFunction(6);
-	return ret;
-}
-
-
 /****************************************************************************
 *	Transmit http message header to destination socket
 *
-*	Parameters:
-*		dsock: destination socket
-*		buf_ctl: buffer for http header
 */
 static int
 xmit_http_message_header(struct socket *dsock,
-			 http_header_buf_ctl_t * buf_ctl)
+			 http_read_ctl_block_t * read_ctl)
 {
-	int ret = 0;
 	struct list_head *l, *temp;
-	http_message_header_buf_t *buf_entry;
-	int len;
+	http_buf_t *buf_entry;
+	int ret = 0;
 
 	EnterFunction(5);
 
 	TCP_VS_DBG(5, "HTTP Message Header:\n");
 
-	list_for_each_safe(l, temp, &buf_ctl->buf_entry_list) {
-		buf_entry =
-		    list_entry(l, http_message_header_buf_t, b_list);
+	list_for_each_safe(l, temp, &read_ctl->buf_entry_list) {
+		buf_entry = list_entry(l, http_buf_t, b_list);
 
-		len = (buf_entry == buf_ctl->cur_buf) ?
-		    buf_ctl->offset : buf_entry->buf_size;
-
-		if (tcp_vs_xmit(dsock, buf_entry->buf, len, MSG_MORE) < 0) {
+		if (tcp_vs_xmit(dsock, buf_entry->buf, 
+				buf_entry->data_len, MSG_MORE) < 0) {
 			TCP_VS_ERR("Error in xmitting message header\n");
 			ret = -1;
 			break;
 		}
-#ifdef CONFIG_TCP_VS_DEBUG
-		*(buf_entry->buf + len - 1) = 0;
-#endif
-		TCP_VS_DBG(5, "%s", buf_entry->buf);
 
-		if (buf_entry != buf_ctl->cur_buf) {
+		if (buf_entry != read_ctl->cur_buf) {
 			list_del(l);
 			free_page((unsigned long) buf_entry->buf);
 			kfree(buf_entry);
 		}
 	}
 
-	buf_ctl->offset = 0;
-	INIT_LIST_HEAD(&buf_ctl->buf_entry_list);
 	LeaveFunction(5);
 	return ret;
 }
 
+/****************************************************************************
+*
+* http_read_reset - reset read buffer
+*
+*/
+int
+http_read_reset (http_read_ctl_block_t *ctl_blk)
+{
+	char* buf = ctl_blk->cur_buf->buf;
+
+	memmove(buf, buf + ctl_blk->offset, ctl_blk->remaining);
+	ctl_blk->offset = 0;
+	return 0;
+}
 
 /****************************************************************************
 *	HTTP content-based scheduling with cookie support:
@@ -822,8 +706,6 @@ tcp_vs_chttp_schedule(struct tcp_vs_conn *conn, struct tcp_vs_service *svc)
 {
 	http_request_t req;
 	http_read_ctl_block_t read_ctl_blk;
-	http_header_buf_ctl_t buf_ctl;
-	char *buffer = NULL;
 	int ret = 1;		/* scheduler has done all the jobs */
 	int len;
 	unsigned long last_read;
@@ -836,35 +718,21 @@ tcp_vs_chttp_schedule(struct tcp_vs_conn *conn, struct tcp_vs_service *svc)
 
 	EnterFunction(5);
 
-	memset(&read_ctl_blk, 0, sizeof(read_ctl_blk));
-	read_ctl_blk.buffer = conn->buffer;
-	read_ctl_blk.len = conn->buflen;
-	read_ctl_blk.flag = 0;
-	read_ctl_blk.sock = conn->csock;
-
 	conn->dest = NULL;
 	conn->dsock = NULL;
 
+	/* init buffer for http message header */
+	if (http_read_init(&read_ctl_blk, conn->csock) != 0) {
+		TCP_VS_ERR("Out of memory!\n");
+		ret = -2;
+		goto out;
+	}
+
 	/* Do we have data ? */
 	while (skb_queue_empty(&(conn->csock->sk->receive_queue))) {
 		interruptible_sleep_on_timeout(&conn->csock->wait, HZ);
 	}
 
-	/* init buffer for http message header */
-	if (http_header_buffer_init(&buf_ctl) != 0) {
-		TCP_VS_ERR("Out of memory!\n");
-		ret = -2;
-		goto out;
-	}
-
-	/* allocate buffer to store data that get from servers */
-	buffer = (char *) get_free_page(GFP_KERNEL);
-	if (buffer == NULL) {
-		TCP_VS_ERR("Out of memory!\n");
-		ret = -2;
-		goto out;
-	}
-
 	last_read = jiffies;
 	do {
 		switch (data_available(&read_ctl_blk)) {
@@ -905,14 +773,13 @@ tcp_vs_chttp_schedule(struct tcp_vs_conn *conn, struct tcp_vs_service *svc)
 		}
 
 		/* read request line from client socket */
-		len = http_read_line(&read_ctl_blk);
+		len = http_read_line(&read_ctl_blk, 1);
 		if (len < 0) {
 			TCP_VS_ERR
 			    ("Error reading request line from client\n");
 			ret = -2;
 			goto out;
 		}
-		write_buffer(&buf_ctl, read_ctl_blk.info, len + 2);
 
 		/* parse the http request line */
 		memset(&req, 0, sizeof(req));
@@ -923,15 +790,12 @@ tcp_vs_chttp_schedule(struct tcp_vs_conn *conn, struct tcp_vs_service *svc)
 			goto out;
 		}
 
-		http_mime_parse_enable(CONTENT_TYPE, 0);
-
 		/* Process MIME header */
 		do {
-			len = http_read_line(&read_ctl_blk);
+			len = http_read_line(&read_ctl_blk, 1);
 			if (len < 0) {
 				goto out;
 			}
-			write_buffer(&buf_ctl, read_ctl_blk.info, len + 2);
 			http_mime_parse(read_ctl_blk.info, len, &req.mime);
 		}
 		while (len != 0);	/* http header end with CRLF,CRLF */
@@ -962,7 +826,7 @@ tcp_vs_chttp_schedule(struct tcp_vs_conn *conn, struct tcp_vs_service *svc)
 			goto lookup_again;
 		}
 
-		if (xmit_http_message_header(dsock, &buf_ctl) != 0) {
+		if (xmit_http_message_header(dsock, &read_ctl_blk) != 0) {
 			goto out_free;
 		}
 
@@ -972,8 +836,8 @@ tcp_vs_chttp_schedule(struct tcp_vs_conn *conn, struct tcp_vs_service *svc)
 			goto out_free;
 		}
 
-		if (chttp_get_response(conn->csock, sc, &req, buffer,
-				       PAGE_SIZE, &close_server) < 0) {
+		if (chttp_get_response(conn->csock, sc, &req, conn->buffer,
+				       conn->buflen, &close_server) < 0) {
 			goto out;
 		}
 
@@ -985,14 +849,13 @@ tcp_vs_chttp_schedule(struct tcp_vs_conn *conn, struct tcp_vs_service *svc)
 		/* have to put back the server connection when
 		   it is not used */
 		tcp_vs_srvconn_put(sc);
+
+		http_read_reset(&read_ctl_blk);
 	}
 	while (req.mime.connection_close != 1);
 
       out:
-	if (buffer) {
-		free_page((unsigned long) buffer);
-	}
-	http_header_buffer_free(&buf_ctl);
+	http_read_free(&read_ctl_blk);
 	LeaveFunction(5);
 	return ret;
 
