@@ -6,7 +6,7 @@
  *
  * tcp_vs_http_parser.c: KTCPVS HTTP parsing engine
  *
- * Version:     $Id: tcp_vs_http_parser.c,v 1.1 2003/05/23 02:08:34 wensong Exp $
+ * Version:     $Id: tcp_vs_http_parser.c,v 1.3 2003/06/14 09:04:11 wensong Exp $
  *
  * Authors:     Wensong Zhang, <wensong@linuxvirtualserver.org>
  *              Hai Long, <david_lung@yahoo.com>
@@ -18,67 +18,38 @@
  *
  */
 
-#include <linux/net.h>
 #include <linux/sched.h>
-#include <linux/skbuff.h>
-#include <net/sock.h>
-#include <asm/uaccess.h>
 #include <linux/ctype.h>
 
 #include "tcp_vs.h"
 
-typedef struct methods {
-	const short number;
-	const char *const name;
-	int len;
-} methods_t;
-
-
 #define HTTP_VERSION_HEADER_LEN		5
 #define HTTP_VERSION_NUMBER_LEN		3
-#define HTTP_VERSION_MAX		3
 
 const char *http_version_header = "HTTP/";
-const char *http_version_number[HTTP_VERSION_MAX] = {
-	"0.9",			/* The position is crucial & magic! do not insert, only add */
-	"1.0",
-	"1.1",
-};
-
-const methods_t http_methods[HTTP_M_MAX] = {
-	{HTTP_M_UNKNOWN, "UNKNOWN", 7},
-	{HTTP_M_OPTIONS, "OPTIONS", 7},
-	{HTTP_M_GET, "GET", 3},
-	{HTTP_M_HEAD, "HEAD", 4},
-	{HTTP_M_POST, "POST", 4},
-	{HTTP_M_PUT, "PUT", 3},
-	{HTTP_M_DELETE, "DELETE", 6},
-	{HTTP_M_TRACE, "TRACE", 5}
-};
 
 #define DEFAULT_MAX_COOKIE_AGE	1800
-
 #define MAX_MIME_HEADER_STRING_LEN	64
 
 typedef void (*HTTP_MIME_PARSER) (http_mime_header_t * mime, char *buffer);
 
-typedef struct {
+typedef struct http_mime_parse_s {
+	struct list_head plist;
 	HTTP_MIME_PARSER parser;
 	char mime_header_string[MAX_MIME_HEADER_STRING_LEN];
-	int mime_header_string_len;
-	int enable;
 } http_mime_parse_t;
 
-static http_mime_parse_t http_mime_parse_table[MAX_PARSER_ID];
+static http_mime_parse_t http_mime_parse_table[MAX_MIME_HEADER_STRING_LEN];
 
 /****************************************************************************
 *	skip whitespace
+*       LWS            = [CRLF] 1*( SP | HT )
 */
 static inline char *
 skip_lws(const char *buffer)
 {
 	char *s = (char *) buffer;
-	while ((*s == ' ') || (*s == '\t') || (*s == '\n') || (*s == '\r')) {
+	while ((*s == ' ') || (*s == '\t')) {
 		s++;
 	}
 	return s;
@@ -87,7 +58,7 @@ skip_lws(const char *buffer)
 /****************************************************************************
 *	search the seperator in a string
 */
-static char *
+char *
 search_sep(const char *s, int len, const char *sep)
 {
 	int l, ll;
@@ -136,8 +107,6 @@ extract_av(char **buffer, char **attribute, char **value)
 		switch (c) {
 		case ' ':
 		case '\t':
-		case '\r':
-		case '\n':
 			end = pos;
 			*end = 0;
 			break;
@@ -176,8 +145,6 @@ extract_av(char **buffer, char **attribute, char **value)
 		switch (c) {
 		case ' ':
 		case '\t':
-		case '\r':
-		case '\n':
 			if ((flag == 0) && (end == NULL)) {
 				end = pos;
 				*end = 0;
@@ -323,7 +290,7 @@ strtol(char *string, char **endptr, int radix)
 *     returned.
 *  In general, any negative number can be considered an overflow error.
 */
-static long
+long
 get_chunk_size(char *b)
 {
 	long chunksize = 0;
@@ -362,6 +329,173 @@ get_chunk_size(char *b)
 	return chunksize;
 }
 
+/****************************************************************************
+ *
+ *  This routine is borrowed from apache server
+ */
+static int
+lookup_builtin_method(const char *method, int len)
+{
+	/* Note: the following code was generated by the "shilka" tool from
+	   the "cocom" parsing/compilation toolkit. It is an optimized lookup
+	   based on analysis of the input keywords. Postprocessing was done
+	   on the shilka output, but the basic structure and analysis is
+	   from there. Should new HTTP methods be added, then manual insertion
+	   into this code is fine, or simply re-running the shilka tool on
+	   the appropriate input. */
+
+	switch (len) {
+	case 3:
+		switch (method[0]) {
+		case 'P':
+			return (method[1] == 'U'
+				&& method[2] == 'T'
+				? HTTP_M_PUT : HTTP_M_UNKNOWN);
+		case 'G':
+			return (method[1] == 'E'
+				&& method[2] == 'T'
+				? HTTP_M_GET : HTTP_M_UNKNOWN);
+		default:
+			return HTTP_M_UNKNOWN;
+		}
+
+	case 4:
+		switch (method[0]) {
+		case 'H':
+			return (method[1] == 'E'
+				&& method[2] == 'A'
+				&& method[3] == 'D'
+				? HTTP_M_GET : HTTP_M_UNKNOWN);
+		case 'P':
+			return (method[1] == 'O'
+				&& method[2] == 'S'
+				&& method[3] == 'T'
+				? HTTP_M_POST : HTTP_M_UNKNOWN);
+		case 'M':
+			return (method[1] == 'O'
+				&& method[2] == 'V'
+				&& method[3] == 'E'
+				? HTTP_M_MOVE : HTTP_M_UNKNOWN);
+		case 'L':
+			return (method[1] == 'O'
+				&& method[2] == 'C'
+				&& method[3] == 'K'
+				? HTTP_M_LOCK : HTTP_M_UNKNOWN);
+		case 'C':
+			return (method[1] == 'O'
+				&& method[2] == 'P'
+				&& method[3] == 'Y'
+				? HTTP_M_COPY : HTTP_M_UNKNOWN);
+		default:
+			return HTTP_M_UNKNOWN;
+		}
+
+	case 5:
+		switch (method[2]) {
+		case 'T':
+			return (memcmp(method, "PATCH", 5) == 0
+				? HTTP_M_PATCH : HTTP_M_UNKNOWN);
+		case 'R':
+			return (memcmp(method, "MERGE", 5) == 0
+				? HTTP_M_MERGE : HTTP_M_UNKNOWN);
+		case 'C':
+			return (memcmp(method, "MKCOL", 5) == 0
+				? HTTP_M_MKCOL : HTTP_M_UNKNOWN);
+		case 'B':
+			return (memcmp(method, "LABEL", 5) == 0
+				? HTTP_M_LABEL : HTTP_M_UNKNOWN);
+		case 'A':
+			return (memcmp(method, "TRACE", 5) == 0
+				? HTTP_M_TRACE : HTTP_M_UNKNOWN);
+		default:
+			return HTTP_M_UNKNOWN;
+		}
+
+	case 6:
+		switch (method[0]) {
+		case 'U':
+			switch (method[5]) {
+			case 'K':
+				return (memcmp(method, "UNLOCK", 6) == 0
+					? HTTP_M_UNLOCK : HTTP_M_UNKNOWN);
+			case 'E':
+				return (memcmp(method, "UPDATE", 6) == 0
+					? HTTP_M_UPDATE : HTTP_M_UNKNOWN);
+			default:
+				return HTTP_M_UNKNOWN;
+			}
+		case 'R':
+			return (memcmp(method, "REPORT", 6) == 0
+				? HTTP_M_REPORT : HTTP_M_UNKNOWN);
+		case 'D':
+			return (memcmp(method, "DELETE", 6) == 0
+				? HTTP_M_DELETE : HTTP_M_UNKNOWN);
+		default:
+			return HTTP_M_UNKNOWN;
+		}
+
+	case 7:
+		switch (method[1]) {
+		case 'P':
+			return (memcmp(method, "OPTIONS", 7) == 0
+				? HTTP_M_OPTIONS : HTTP_M_UNKNOWN);
+		case 'O':
+			return (memcmp(method, "CONNECT", 7) == 0
+				? HTTP_M_CONNECT : HTTP_M_UNKNOWN);
+		case 'H':
+			return (memcmp(method, "CHECKIN", 7) == 0
+				? HTTP_M_CHECKIN : HTTP_M_UNKNOWN);
+		default:
+			return HTTP_M_UNKNOWN;
+		}
+
+	case 8:
+		switch (method[0]) {
+		case 'P':
+			return (memcmp(method, "PROPFIND", 8) == 0
+				? HTTP_M_PROPFIND : HTTP_M_UNKNOWN);
+		case 'C':
+			return (memcmp(method, "CHECKOUT", 8) == 0
+				? HTTP_M_CHECKOUT : HTTP_M_UNKNOWN);
+		default:
+			return HTTP_M_UNKNOWN;
+		}
+
+	case 9:
+		return (memcmp(method, "PROPPATCH", 9) == 0
+			? HTTP_M_PROPPATCH : HTTP_M_UNKNOWN);
+
+	case 10:
+		switch (method[0]) {
+		case 'U':
+			return (memcmp(method, "UNCHECKOUT", 10) == 0
+				? HTTP_M_UNCHECKOUT : HTTP_M_UNKNOWN);
+		case 'M':
+			return (memcmp(method, "MKACTIVITY", 10) == 0
+				? HTTP_M_MKACTIVITY : HTTP_M_UNKNOWN);
+		default:
+			return HTTP_M_UNKNOWN;
+		}
+
+	case 11:
+		return (memcmp(method, "MKWORKSPACE", 11) == 0
+			? HTTP_M_MKWORKSPACE : HTTP_M_UNKNOWN);
+
+	case 15:
+		return (memcmp(method, "VERSION-CONTROL", 15) == 0
+			? HTTP_M_VERSION_CONTROL : HTTP_M_UNKNOWN);
+
+	case 16:
+		return (memcmp(method, "BASELINE-CONTROL", 16) == 0
+			? HTTP_M_BASELINE_CONTROL : HTTP_M_UNKNOWN);
+
+	default:
+		return HTTP_M_UNKNOWN;
+	}
+
+	/* NOTREACHED */
+}
+
 
 /****************************************************************************
 *   Parse http request line. (request line is terminated by CRLF)
@@ -374,17 +508,18 @@ get_chunk_size(char *b)
 *
 */
 int
-parse_http_request_line(char *buffer, size_t len, http_request_t * req)
+parse_http_request_line(char *buffer, size_t size, http_request_t * req)
 {
-	char *pos, c;
+	char *pos, *method, *ver;
+	char c;
 	int ret = PARSE_ERROR;
-	int i;
+	int len, major, minor;
 
 	EnterFunction(5);
 
 	/* terminate string */
-	c = buffer[len];
-	buffer[len] = 0;
+	c = buffer[size];
+	buffer[size] = 0;
 
 	TCP_VS_DBG(5, "parsing request:\n");
 	TCP_VS_DBG(5, "--------------------\n");
@@ -392,7 +527,7 @@ parse_http_request_line(char *buffer, size_t len, http_request_t * req)
 	TCP_VS_DBG(5, "--------------------\n");
 
 	req->message = buffer;
-	req->message_len = len;
+	req->message_len = size;
 
 	/*
 	 * RFC 2616, 5.1:
@@ -400,58 +535,49 @@ parse_http_request_line(char *buffer, size_t len, http_request_t * req)
 	 */
 
 	/* try to get method */
-	pos = skip_lws(buffer);
-	req->method = HTTP_M_UNKNOWN;	/* Default :) */
-	for (i = 1; i < HTTP_M_MAX; i++) {
-		if (strnicmp
-		    (pos, http_methods[i].name, http_methods[i].len) == 0) {
-			req->method = i;
-			break;
-		}
-	}
-	if (req->method == HTTP_M_UNKNOWN) {
+	method = skip_lws(buffer);
+	if ((pos = strchr((char *) method, SP)) == NULL) {
 		goto exit;
 	}
-	TCP_VS_DBG(6, "HTTP METHOD: %s\n", http_methods[i].name);
+	len = pos - method;
 
-	pos += http_methods[i].len;
+	req->method = lookup_builtin_method(method, len);
+	if (req->method == HTTP_M_UNKNOWN) {
+		TCP_VS_ERR("Unknow http method.\n");
+		goto exit;
+	}
 
 	/* get URI string */
 	req->uri_str = skip_lws(pos + 1);
-	TCP_VS_DBG(6, "URI: %s\n", req->uri_str);
-
 	if ((pos = strchr((char *) req->uri_str, SP)) == NULL) {
 		goto exit;
 	}
-
 	req->uri_len = pos - req->uri_str;
 
 	/* get http version */
 	req->version_str = skip_lws(pos + 1);
+
 	if (strnicmp(req->version_str,
 		     http_version_header, HTTP_VERSION_HEADER_LEN) != 0) {
 		goto exit;
 	}
 
-	req->version = HTTP_V_UNKNOWN;
-	req->version_str += HTTP_VERSION_HEADER_LEN;
-	for (i = HTTP_VERSION_MAX - 1; i >= 0; i--) {
-		if (strncmp(req->version_str,
-			    http_version_number[i],
-			    HTTP_VERSION_NUMBER_LEN) == 0) {
-			req->version = i + HTTP_V_0_9;
-			break;
-		}
-	}
+	ver = (char *) req->version_str + HTTP_VERSION_HEADER_LEN;
+	len = strlen(ver);
 
-	if (req->version == HTTP_V_UNKNOWN) {
-		goto exit;
-	}
+	/* Avoid sscanf in the common case */
+	if (len == HTTP_VERSION_NUMBER_LEN
+	    && isdigit(ver[0]) && ver[1] == '.' && isdigit(ver[2])) {
+		req->version = HTTP_VERSION(ver[0] - '0', ver[2] - '0');
+	} else if (2 == sscanf(ver, "%u.%u", &major, &minor)
+		   && (minor < HTTP_VERSION(1, 0)))	/* don't allow HTTP/0.1000 */
+		req->version = HTTP_VERSION(major, minor);
+	else
+		req->version = HTTP_VERSION(1, 0);
 
-	TCP_VS_DBG(6, "HTTP VERSION: %s\n", http_version_number[i]);
 	ret = PARSE_OK;
       exit:
-	buffer[len] = c;	/* restore string */
+	buffer[size] = c;	/* restore string */
 	LeaveFunction(5);
 	return ret;
 }
@@ -468,18 +594,19 @@ parse_http_request_line(char *buffer, size_t len, http_request_t * req)
 *
 */
 int
-parse_http_status_line(char *buffer, size_t len, http_response_t * resp)
+parse_http_status_line(char *buffer, size_t size, http_response_t * resp)
 {
-	char *pos, c;
-	int i, ret = PARSE_ERROR;
+	char *pos, *ver;
+	char c;
+	int len, major, minor;
 
 	EnterFunction(5);
 
 	assert(buffer != NULL);
 
 	/* terminate string */
-	c = buffer[len];
-	buffer[len] = '\0';
+	c = buffer[size];
+	buffer[size] = '\0';
 
 	TCP_VS_DBG(5, "parsing response:\n");
 	TCP_VS_DBG(5, "--------------------\n");
@@ -491,36 +618,39 @@ parse_http_status_line(char *buffer, size_t len, http_response_t * resp)
 	 *      Status-Line = HTTP-Version SP Status-Code SP Reason-Phrase CRLF
 	 */
 
-	pos = skip_lws(buffer);
+	ver = skip_lws(buffer);
+	if ((pos = strchr((char *) ver, SP)) == NULL) {
+		return PARSE_ERROR;
+	}
 
-	if (strnicmp(pos,
+	if (strnicmp(ver,
 		     http_version_header, HTTP_VERSION_HEADER_LEN) != 0) {
-		return ret;
+		return PARSE_ERROR;
 	}
 
-	pos += HTTP_VERSION_HEADER_LEN;
-	resp->version = HTTP_V_UNKNOWN;
-	for (i = HTTP_VERSION_MAX - 1; i >= 0; i--) {
-		if (strnicmp(pos,
-			     http_version_number[i],
-			     HTTP_VERSION_NUMBER_LEN) == 0) {
-			resp->version = i + HTTP_V_0_9;
-			break;
-		}
-	}
+	ver += HTTP_VERSION_HEADER_LEN;
+	len = pos - ver;
 
-	pos += HTTP_VERSION_NUMBER_LEN;
-	if (resp->version != HTTP_V_UNKNOWN) {
-		/* get the status code */
-		resp->status_code = strtol(pos, NULL, 10);
-		assert(resp->status_code >= 100);
-		TCP_VS_DBG(6, "Status Code: %d\n", resp->status_code);
-		ret = PARSE_OK;
-	}
+	/* Avoid sscanf in the common case */
+	if (len == HTTP_VERSION_NUMBER_LEN
+	    && isdigit(ver[0]) && ver[1] == '.' && isdigit(ver[2])) {
+		resp->version = HTTP_VERSION(ver[0] - '0', ver[2] - '0');
+	} else if (2 == sscanf(ver, "%u.%u", &major, &minor)
+		   && (minor < HTTP_VERSION(1, 0)))	/* don't allow HTTP/0.1000 */
+		resp->version = HTTP_VERSION(major, minor);
+	else
+		resp->version = HTTP_VERSION(1, 0);
 
-	buffer[len] = c;	/* restore string */
+
+	/* get the status code */
+	resp->status_code = strtol(pos + 1, NULL, 10);
+	assert(resp->status_code >= 100);
+	TCP_VS_DBG(6, "Status Code: %d\n", resp->status_code);
+
+	buffer[size] = c;	/* restore string */
 	LeaveFunction(5);
-	return ret;
+
+	return PARSE_OK;
 }
 
 
@@ -586,29 +716,39 @@ http_line_unescape(char *string,	/* escaped string to unescape */
 * register_mime_parser - register a http mime header parser
 *
 */
-static void register_mime_parser
-    (int parser_id, HTTP_MIME_PARSER mime_parser, char *mime_header_string) {
-	assert(parser_id < MAX_PARSER_ID);
-	assert(mime_parser != NULL);
-	assert(mime_header_string != NULL);
-
-	strncpy(http_mime_parse_table[parser_id].mime_header_string,
-		mime_header_string, MAX_MIME_HEADER_STRING_LEN);
-	http_mime_parse_table[parser_id].mime_header_string_len =
-	    strlen(mime_header_string);
-	http_mime_parse_table[parser_id].parser = mime_parser;
-	http_mime_parse_table[parser_id].enable = 1;
-}
-
-/****************************************************************************
-*
-* http_mime_parse_enable - enable or disable a specified http mime header parser
-*
-*/
-void
-http_mime_parse_enable(int parser_id, int enable)
+static void
+register_mime_parser(HTTP_MIME_PARSER parser, char *mime_str)
 {
-	http_mime_parse_table[parser_id].enable = enable;
+	int index;
+	http_mime_parse_t *parse_entry, *new_entry;
+
+	assert(parser != NULL);
+	assert(mime_str != NULL);
+
+	index = strlen(mime_str);
+	if (index >= MAX_MIME_HEADER_STRING_LEN)
+		return;
+
+	parse_entry = &http_mime_parse_table[index];
+	if (parse_entry->parser == NULL) {
+		INIT_LIST_HEAD(&parse_entry->plist);
+	} else {
+		if (strnicmp
+		    (mime_str, parse_entry->mime_header_string,
+		     index) == 0) {
+			return;
+		}
+		new_entry = kmalloc(sizeof(http_mime_parse_t), GFP_KERNEL);
+		if (new_entry == NULL) {
+			TCP_VS_ERR("Out of memory!\n");
+			return;
+		}
+		list_add_tail(&new_entry->plist, &parse_entry->plist);
+		parse_entry = new_entry;
+	}
+
+	strcpy(parse_entry->mime_header_string, mime_str);
+	parse_entry->parser = parser;
 }
 
 /****************************************************************************
@@ -621,7 +761,7 @@ transfer_encoding_parser(http_mime_header_t * mime, char *buffer)
 {
 	EnterFunction(6);
 
-	if (strnicmp(buffer, "identity", 8) != 0) {
+	if (strnicmp(buffer, "chunked", 7) == 0) {
 		mime->transfer_encoding = 1;
 		TCP_VS_DBG(6, "Transfer-Encoding: chunked\n");
 	}
@@ -729,17 +869,18 @@ content_type_parser(http_mime_header_t * mime, char *buffer)
 *                   |       "Secure"
 *                   |       "Version" "=" 1*DIGIT*
 *
-*	Note: the input buffer will be modified indirectly.
 */
 static void
-set_cookie_parser(http_mime_header_t * mime, char *buffer)
+set_cookie_parser(http_mime_header_t * mime, char *buf)
 {
 	http_cookie_t *ck;
+	char* buffer;
 	char *attribute, *value, *s;
 	int r;
 
 	EnterFunction(6);
 
+	buffer = strdup(buf);
 	TCP_VS_DBG(5, "Set-Cookie:%s", buffer);
 
 	mime->set_cookie2 = 0;
@@ -779,8 +920,6 @@ set_cookie_parser(http_mime_header_t * mime, char *buffer)
 		list_add_tail(&ck->c_list, &mime->cookie_list);
 
 		ck->key->name = strdup(attribute);
-		ck->key->value = strdup(value);
-
 		if (r == 0) {
 			goto out;
 		}
@@ -811,6 +950,7 @@ set_cookie_parser(http_mime_header_t * mime, char *buffer)
 	}			/* end for */
 
       out:
+	kfree(buffer);
 	LeaveFunction(6);
 	return;
 }
@@ -836,17 +976,18 @@ set_cookie_parser(http_mime_header_t * mime, char *buffer)
 *   portlist        =       1#portnum
 *   portnum         =       1*DIGIT
 *
-*	Note: the input buffer will be modified indirectly.
 */
 static void
-set_cookie2_parser(http_mime_header_t * mime, char *buffer)
+set_cookie2_parser(http_mime_header_t * mime, char *buf)
 {
 	http_cookie_t *ck;
 	char *attribute, *value, *s;
+	char *buffer;
 	int r;
 
 	EnterFunction(6);
 
+	buffer = strdup(buf);
 	TCP_VS_DBG(5, "Set-Cookie2:%s", buffer);
 
 	mime->set_cookie2 = 1;
@@ -886,8 +1027,6 @@ set_cookie2_parser(http_mime_header_t * mime, char *buffer)
 		list_add_tail(&ck->c_list, &mime->cookie_list);
 
 		ck->key->name = strdup(attribute);
-		ck->key->value = strdup(value);
-
 		if (r == 0) {
 			goto out;
 		}
@@ -920,6 +1059,7 @@ set_cookie2_parser(http_mime_header_t * mime, char *buffer)
 	}			/* end for */
 
       out:
+	kfree(buffer);
 	LeaveFunction(6);
 	return;
 }
@@ -941,13 +1081,15 @@ set_cookie2_parser(http_mime_header_t * mime, char *buffer)
 *	      omitted.
 */
 static void
-cookie_parser(http_mime_header_t * mime, char *buffer)
+cookie_parser(http_mime_header_t * mime, char *buf)
 {
 	char *pos, *attribute, *value;
+	char* buffer;
 	int r = 2;
 
 	EnterFunction(6);
 
+	buffer = strdup(buf);
 	TCP_VS_DBG(5, "\nCookie:%s ", buffer);
 
 	pos = skip_lws(buffer);
@@ -964,6 +1106,7 @@ cookie_parser(http_mime_header_t * mime, char *buffer)
 		}
 	}
 
+	kfree(buffer);
 	LeaveFunction(6);
 	return;
 }
@@ -972,17 +1115,13 @@ void
 http_mime_parser_init(void)
 {
 	memset(http_mime_parse_table, 0, sizeof(http_mime_parse_table));
-	register_mime_parser(TRANSFER_ENCODING, transfer_encoding_parser,
-			     "Transfer-Encoding");
-	register_mime_parser(CONTENT_LENGTH, content_length_parser,
-			     "Content-Length");
-	register_mime_parser(CONNECTION, connection_parser, "Connection");
-	register_mime_parser(CONTENT_TYPE, content_type_parser,
-			     "Content-type");
-	register_mime_parser(SET_COOKIE, set_cookie_parser, "Set-Cookie");
-	register_mime_parser(SET_COOKIE2, set_cookie2_parser,
-			     "Set-Cookie2");
-	register_mime_parser(COOKIE, cookie_parser, "Cookie");
+	register_mime_parser(transfer_encoding_parser, "Transfer-Encoding");
+	register_mime_parser(content_length_parser, "Content-Length");
+	register_mime_parser(connection_parser, "Connection");
+	register_mime_parser(content_type_parser, "Content-type");
+	register_mime_parser(set_cookie_parser, "Set-Cookie");
+	register_mime_parser(set_cookie2_parser, "Set-Cookie2");
+	register_mime_parser(cookie_parser, "Cookie");
 }
 
 /******************************************************************************
@@ -997,9 +1136,8 @@ int
 http_mime_parse(char *buffer, int len, http_mime_header_t * mime)
 {
 	char *pos, c;
-	int i, l, ret = PARSE_ERROR;
-
-	EnterFunction(5);
+	int l, ret = PARSE_OK;
+	http_mime_parse_t *parse_entry, *pe;
 
 	assert(buffer != NULL);
 
@@ -1011,488 +1149,43 @@ http_mime_parse(char *buffer, int len, http_mime_header_t * mime)
 
 	buffer = skip_lws(buffer);
 	if ((pos = strchr(buffer, ':')) == NULL) {
-		return PARSE_ERROR;
+		ret = PARSE_ERROR;
+		goto exit;
 	}
 
 	l = pos - buffer;
+	if (l >= MAX_MIME_HEADER_STRING_LEN)
+		goto exit;
+
+	parse_entry = &http_mime_parse_table[l];
+
+	if (parse_entry->parser == NULL) {	/* an unregistered mime header */
+		goto exit;
+	}
+
+	assert(parse_entry->mime_header_string != NULL);
+
+	if (strnicmp(parse_entry->mime_header_string, buffer, l) != 0) {
+		struct list_head *list;
+		int found = 0;
+		list_for_each(list, &parse_entry->plist) {
+			pe = list_entry(list, http_mime_parse_t, plist);
+			if (strnicmp(pe->mime_header_string, buffer, l) ==
+			    0) {
+				parse_entry = pe;
+				found = 1;
+				break;
+			}
+		}
+		if (!found)	/* an unregistered mime header */
+			goto exit;
+	}
+
 	pos = skip_lws(pos + 1);
+	parse_entry->parser(mime, pos);
 
-	for (i = 0; i < MAX_PARSER_ID; i++) {
-		if (http_mime_parse_table[i].enable &&
-		    (l == http_mime_parse_table[i].mime_header_string_len)
-		    &&
-		    (strnicmp
-		     (http_mime_parse_table[i].mime_header_string, buffer,
-		      l) == 0)) {
-			http_mime_parse_table[i].parser(mime, pos);
-			break;
-		}
-	}
-
+      exit:
 	buffer[len] = c;	/* restore string */
-	LeaveFunction(5);
 	return ret;
 }
 
-/****************************************************************************
-*	relay data between source socket and destination socket
-*/
-static int
-relay_http_data(struct socket *dsock,	/* destination socket */
-		http_read_ctl_block_t * ctl_blk,	/* read control block with source socket */
-		int len		/* relay data length */
-    )
-{
-	int nbytes, reads, w = 0;
-	int ret = -1;
-
-	DECLARE_WAITQUEUE(wait, current);
-
-	EnterFunction(5);
-
-	assert(ctl_blk->remaining <= (ctl_blk->len - ctl_blk->offset));
-	assert(len > 0);
-
-	/* if there is enough data in read buffer */
-	nbytes = len - ctl_blk->remaining;
-	if (nbytes <= 0) {
-		if (tcp_vs_xmit(dsock, ctl_blk->buffer + ctl_blk->offset,
-				len, MSG_MORE) < 0) {
-			TCP_VS_ERR("Error in xmitting message body\n");
-			goto exit;
-		}
-		ctl_blk->offset += len;
-		ctl_blk->remaining -= len;
-		goto done;
-	}
-
-	/* xmit the remaining bytes */
-	if (ctl_blk->remaining > 0) {
-		if (tcp_vs_xmit(dsock, ctl_blk->buffer + ctl_blk->offset,
-				ctl_blk->remaining, MSG_MORE) < 0) {
-			TCP_VS_ERR("Error in xmitting remaining bytes\n");
-			goto exit;
-		}
-	}
-
-	do {
-		reads =
-		    tcp_vs_recvbuffer(ctl_blk->sock, ctl_blk->buffer,
-				      ctl_blk->len, ctl_blk->flag);
-		if (reads == 0) {
-			TCP_VS_DBG(5, "Reads 0 bytes while relay\n");
-			add_wait_queue(ctl_blk->sock->sk->sleep, &wait);
-			__set_current_state(TASK_INTERRUPTIBLE);
-			schedule();
-			__set_current_state(TASK_RUNNING);
-			remove_wait_queue(ctl_blk->sock->sk->sleep, &wait);
-			continue;
-		}
-
-		if (reads < 0) {
-			TCP_VS_ERR("Error in reading while relaying\n");
-			goto exit;
-		}
-
-		w = MIN(nbytes, reads);
-
-		if (tcp_vs_xmit(dsock, ctl_blk->buffer, w, MSG_MORE) < 0) {
-			TCP_VS_ERR("Error in relaying bytes\n");
-			goto exit;
-		}
-
-		nbytes -= w;
-	} while (nbytes > 0);
-
-	ctl_blk->offset = w;
-	ctl_blk->remaining = reads - w;
-
-	assert(ctl_blk->remaining >= 0);
-	assert(ctl_blk->offset < ctl_blk->len);
-
-      done:
-	ret = 0;
-      exit:
-	LeaveFunction(5);
-	return ret;
-}
-
-/****************************************************************************
-*
-* http_read_line - read a line from socket.
-*
-*   At first, get the line from the remaining bytes. then read bytes without
-* move the buffer. finally, move the memory and read a full buffer to get a
-* line.
-*   Return the len of the line (not including CRLF), or -1 if failed.
-*
-*   Note:
-*	1, http_read_line does not terminate the line with '\0', it still end
-*   with CRLF.
-*	2, If the line length is larger than the size of the buffer, it will
-*   failed.
-*
-*/
-int
-http_read_line(http_read_ctl_block_t * ctl_blk)
-{
-	char *buf;
-	int nbytes, i, offset, reads, move;
-	int len = -1;
-
-	DECLARE_WAITQUEUE(wait, current);
-
-	EnterFunction(5);
-
-	assert(ctl_blk->remaining <= (ctl_blk->len - ctl_blk->offset));
-
-	ctl_blk->info = NULL;
-	if (ctl_blk->remaining == 0) {
-		ctl_blk->offset = 0;
-	}
-
-	offset = ctl_blk->offset;
-	buf = ctl_blk->buffer + offset;
-
-	/* try to get a line from the remaining bytes */
-	for (i = 0; i < ctl_blk->remaining - 1; i++) {
-		if ((buf[i] == CR) && (buf[i + 1] == LF)) {
-			len = i;
-			goto done;
-		}
-	}
-
-	move = 0;
-
-      get_a_line:
-	nbytes = ctl_blk->len - ctl_blk->offset - ctl_blk->remaining;
-
-	/* try to read a line from the socket */
-	while ((nbytes > 0) && (len < 0)) {
-		/* go out if the connection is closed */
-		if (ctl_blk->sock->sk->state != TCP_ESTABLISHED
-		    && ctl_blk->sock->sk->state != TCP_CLOSE_WAIT) {
-			if (len > 0)
-				goto done;
-			else
-				goto exit;
-		}
-
-		assert(ctl_blk->remaining <=
-		       (ctl_blk->len - ctl_blk->offset));
-		reads =
-		    tcp_vs_recvbuffer(ctl_blk->sock,
-				      ctl_blk->buffer + ctl_blk->offset +
-				      ctl_blk->remaining,
-				      ctl_blk->len - ctl_blk->offset -
-				      ctl_blk->remaining, ctl_blk->flag);
-
-		if (reads == 0) {
-			TCP_VS_DBG(5,
-				   "Read 0 bytes while reading a line\n");
-			add_wait_queue(ctl_blk->sock->sk->sleep, &wait);
-			__set_current_state(TASK_INTERRUPTIBLE);
-			schedule_timeout(HZ);
-			__set_current_state(TASK_RUNNING);
-			remove_wait_queue(ctl_blk->sock->sk->sleep, &wait);
-			continue;
-		}
-
-		if (reads < 0) {
-			TCP_VS_ERR("Error in reading a line\n");
-			goto exit;
-		}
-
-		ctl_blk->remaining += reads;
-
-		/* try to get a line from the remaing bytes */
-		for (; i < ctl_blk->remaining - 1; i++) {
-			if ((buf[i] == CR) && (buf[i + 1] == LF)) {
-				len = i;
-				goto done;
-			}
-		}
-		nbytes -= reads;
-	}
-
-	/* memmove and read again */
-	if ((len < 0) && (move == 0)) {
-		memmove(ctl_blk->buffer, buf, ctl_blk->remaining);
-		ctl_blk->offset = 0;
-		buf = ctl_blk->buffer;
-		move = 1;
-		goto get_a_line;
-	} else {
-		assert((len < 0) && (move == 1));
-		TCP_VS_ERR("Buffer is too small while reading a line.\n");
-		goto exit;
-	}
-
-      done:
-	ctl_blk->info = buf;
-	ctl_blk->offset += len + 2;
-	ctl_blk->remaining -= len + 2;
-
-	assert(ctl_blk->remaining >= 0);
-	assert(ctl_blk->offset < ctl_blk->len);
-
-      exit:
-	LeaveFunction(5);
-	return len;
-}
-
-
-/****************************************************************************
-* relay_multiparts: relay multipart/byteranges body
-*
-*  relay all data until "CRLF--THIS_STRING_SEPARATES--CRLF" is found.
-*
-*  Note: there may be a endless loop if the separate string is not found, tbd.
-*
-*/
-static int relay_multiparts
-    (struct socket *dsock,
-     http_read_ctl_block_t * ctl_blk, http_mime_header_t * mime) {
-	int len, sep_len, l, reads;
-	int ret = -1;
-	char *buf, *pos;
-	char *sep = NULL;
-
-	DECLARE_WAITQUEUE(wait, current);
-
-	EnterFunction(5);
-
-	sep_len = strlen(mime->sep) + 8;
-	if ((sep = kmalloc(sep_len + 1, GFP_KERNEL)) == NULL) {
-		goto exit;
-	}
-
-	snprintf(sep, sep_len + 1, "\r\n--%s--\r\n", mime->sep);
-
-	/* deal with the remaining bytes */
-	buf = ctl_blk->buffer + ctl_blk->offset;
-	len = ctl_blk->remaining;
-
-	if ((len > 0) && (tcp_vs_xmit(dsock, buf, len, MSG_MORE) < 0)) {
-		TCP_VS_ERR("Error in xmitting multiparts (remaining)\n");
-		goto exit;
-	}
-
-	pos = search_sep(buf, len, sep);
-	if (pos != NULL) {
-		goto done;
-	}
-
-	l = MIN(len, sep_len);
-	memmove(ctl_blk->buffer, buf + len - l, l);
-
-	/* search for CRLF--THIS_STRING_SEPARATES--CRLF */
-	while (1) {
-		reads =
-		    tcp_vs_recvbuffer(ctl_blk->sock, ctl_blk->buffer + l,
-				      ctl_blk->len - l, 0);
-		if (reads == 0) {
-			TCP_VS_DBG(5,
-				   "Reads 0 bytes while relaying multiparts\n");
-			add_wait_queue(ctl_blk->sock->sk->sleep, &wait);
-			__set_current_state(TASK_INTERRUPTIBLE);
-			schedule();
-			__set_current_state(TASK_RUNNING);
-			remove_wait_queue(ctl_blk->sock->sk->sleep, &wait);
-			continue;
-		}
-
-		if (reads < 0) {
-			TCP_VS_ERR("Error in receiving multiparts\n");
-			goto exit;
-		}
-
-		if (tcp_vs_xmit
-		    (dsock, ctl_blk->buffer + l, reads, MSG_MORE) < 0) {
-			TCP_VS_ERR("Error in xmitting multiparts\n");
-			goto exit;
-		}
-
-		len = l + reads;
-		pos = search_sep(ctl_blk->buffer, len, sep);
-		if (pos != NULL) {
-			goto done;
-		}
-
-		l = MIN(len, sep_len);
-		memmove(ctl_blk->buffer, ctl_blk->buffer + len - l, l);
-	}
-
-      done:
-	ret = 0;
-      exit:
-	if (sep) {
-		kfree(sep);
-	}
-
-	LeaveFunction(5);
-	return ret;
-}
-
-/****************************************************************************
-* transfer http message body.
-*
-* When a message-body is included with a message, the transfer-length of that
-* body is determined by one of the following (in order of precedence):
-*  1. Any response message which "MUST NOT" include a message-body (such as
-* the 1xx, 204, and 304 responses and any response to a HEAD request) is always
-* terminated by the first empty line after the header fields, regardless of the
-* entity-header fields present in the message.
-*  2. If a Transfer-Encoding header field (section 14.41) is present and has
-* any value other than "identity", then the transfer-length is defined by use of
-* the "chunked" transfer-coding (section 3.6), unless the message is terminated
-* by closing the connection.
-*  3. If a Content-Length header field (section 14.13) is present, its decimal
-* value in OCTETs represents both the entity-length and the transfer-length. The
-* Content-Length header field MUST NOT be sent if these two lengths are different
-* (i.e., if a Transfer-Encoding header field is present). If a message is received
-* with both a Transfer-Encoding header field and a Content-Length header field,
-* the latter MUST be ignored.
-*  4. If the message uses the media type "multipart/byteranges", and the transfer-
-* length is not otherwise specified, then this self-delimiting media type defines
-* the transfer-length. This media type MUST NOT be used unless the sender knows
-* that the recipient can arse it; the presence in a request of a Range header with
-* multiple byte-range specifiers from a 1.1 client implies that the client can parse
-* multipart/byteranges responses.
-*      A range header might be forwarded by a 1.0 proxy that does not
-	understand multipart/byteranges; in this case the server MUST
-	delimit the message using methods defined in items 1,3 or 5 of this
-	section.
-*  5. By the server closing the connection. (Closing the connection cannot be
-* used to indicate the end of a request body, since that would leave no possibility
-* for the server to send back a response.)
-*
-*/
-int
-relay_http_message_body(struct socket *dsock,	/* destination socket */
-			http_read_ctl_block_t * ctl_blk,	/* read control block */
-			http_mime_header_t * mime)
-{
-	int ret = -1;
-
-	EnterFunction(5);
-
-	if (mime->transfer_encoding) {
-		/*
-		 * 19.4.6 Introduction of Transfer-Encoding
-		 *           HTTP/1.1 introduces the Transfer-Encoding header field (section
-		 *           14.41). Proxies/gateways MUST remove any transfer-coding prior to
-		 *        forwarding a message via a MIME-compliant protocol.
-		 *           A process for decoding the "chunked" transfer-coding (section 3.6) can be
-		 *        represented in pseudo-code as:
-		 *               length := 0
-		 *               read chunk-size, chunk-extension (if any) and CRLF
-		 *               while (chunk-size > 0) {
-		 *                      read chunk-data and CRLF
-		 *                      append chunk-data to entity-body
-		 *                      length := length + chunk-size
-		 *                      read chunk-size and CRLF
-		 *               }
-		 *               read entity-header
-		 *               while (entity-header not empty) {
-		 *                      append entity-header to existing header fields
-		 *                      read entity-header
-		 *               }
-		 *               Content-Length := length
-		 *               Remove "chunked" from Transfer-Encoding
-		 */
-		int len, chunk_size;
-		do {
-			len = http_read_line(ctl_blk);
-			if (len < 0) {
-				TCP_VS_ERR
-				    ("Error in reading chunk size from client\n");
-				goto exit;
-			}
-
-			if (tcp_vs_xmit
-			    (dsock, ctl_blk->info, len + 2, MSG_MORE) < 0) {
-				TCP_VS_ERR
-				    ("Error in xmitting chunk size & extension\n");
-				goto exit;
-			}
-
-			ctl_blk->info[len] = 0;
-			chunk_size = get_chunk_size(ctl_blk->info);
-
-			TCP_VS_DBG(5, "Chunked line: %s\n", ctl_blk->info);
-
-			if (chunk_size > 0) {
-				if (relay_http_data
-				    (dsock, ctl_blk, chunk_size + 2) < 0) {
-					TCP_VS_ERR
-					    ("Error in xmitting chunk data\n");
-					goto exit;
-				}
-			}
-		} while (chunk_size > 0);
-
-		/* relay the trailer */
-		do {
-			len = http_read_line(ctl_blk);
-			if (len < 0) {
-				TCP_VS_ERR("Error in reading trailer.\n");
-				goto exit;
-			}
-			if (tcp_vs_xmit
-			    (dsock, ctl_blk->info, len + 2, MSG_MORE) < 0) {
-				TCP_VS_ERR("Error in xmitting trailer\n");
-				goto exit;
-			}
-		} while (len != 0);
-		ret = 0;
-	} else if (mime->content_length) {
-		ret =
-		    relay_http_data(dsock, ctl_blk, mime->content_length);
-	} else if (mime->sep) {
-		ret = relay_multiparts(dsock, ctl_blk, mime);
-	} else {
-		ret = 0;	/* ? */
-	}
-
-      exit:
-	LeaveFunction(5);
-	return ret;
-}
-
-/****************************************************************************
-*	Is there any data in socket?
-*
-* return:
-*	-1,	Socket error
-*	 0,	No data can read from socket
-*	 1,	Data available
-*/
-int
-data_available(http_read_ctl_block_t * ctl_blk)
-{
-	int ret;
-
-	EnterFunction(12);
-
-	if (ctl_blk->remaining == 0) {
-		/* check if the connection is closed */
-		if (ctl_blk->sock->sk->state != TCP_ESTABLISHED
-		    && ctl_blk->sock->sk->state != TCP_CLOSE_WAIT) {
-			ret = -1;
-			goto out;
-		}
-
-		/* Do we have data ? */
-		if (skb_queue_empty(&(ctl_blk->sock->sk->receive_queue)))
-			ret = 0;
-		else
-			ret = 1;
-		goto out;
-	} else
-		ret = 1;
-
-      out:
-	LeaveFunction(12);
-	return ret;
-}
